// file      : shared-base.hxx
// author    : Boris Kolpackov <boris@kolpackov.net>
// copyright : Copyright (c) 2008 Boris Kolpackov
// license   : GNU GPL v2; see accompanying LICENSE file

#ifndef SHARED_BASE_HXX
#define SHARED_BASE_HXX

#include <new>
#include <cstddef>   // std::size_t
#include <exception> // std::exception


struct share
{
  explicit
  share (char id);

  bool
  operator== (share) const;

private:
  char id_;
};

extern share shared;
extern share exclusive;

void*
operator new (std::size_t, share) throw (std::bad_alloc);

void
operator delete (void*, share) throw ();

struct not_shared: std::exception
{
  virtual char const*
  what () const throw ();
};

struct shared_base
{
  shared_base ();
  shared_base (shared_base const&);
  shared_base&
  operator= (shared_base const&);

  void
  _inc_ref ();

  bool
  _dec_ref ();

  std::size_t
  _ref_count () const;

  void*
  operator new (std::size_t, share) throw (std::bad_alloc);

  void
  operator delete (void*, share) throw ();

  void
  operator delete (void*) throw ();

protected:
  std::size_t counter_;
};

template <typename X>
inline X*
inc_ref (X*);

template <typename X>
inline void
dec_ref (X*);

template <typename X>
inline std::size_t
ref_count (X const*);

#include "shared-base.ixx"
#include "shared-base.txx"

#endif // SHARED_PTR_HXX
